////////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2001-2023 Alexey Kuryakin daqgroup@mail.ru under MIT license //
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// This file is part of the CRW-DAQ project by DaqGroup - component CRWLIB.   //
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Purpose:                                                                   //
// Wrapper for using shared memory manager.                                   //
// NB: ALWAYS use this unit as first unit in USES clause list.                //
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// History:                                                                   //
// 20230426 - Created, A.K.                                                   //
// 20240324 - ExtMM,ExtMM_Allocated                                              //
////////////////////////////////////////////////////////////////////////////////

unit _crw_sharemem; //  Wrapper for useing Shared Memory Manager.

{$I _crw_sysdef}

{$IFNDEF IS_EXE_UNIT}
 {$DEFINE IS_DLL_UNIT}
{$ENDIF}

{$IFDEF FPC}
 {$MODE Delphi}{$H+}
{$ENDIF}

{$WARN 5023 off : Unit "$1" not used in $2}

interface

uses
 {$IFDEF UNIX}
  cthreads, dl,
   {$IFDEF USES_SHAREMEM}
    {$IFNDEF IS_DLL_UNIT}
     {$IFDEF USES_SYNFPCX64MM} synfpcx64mm, {$ENDIF}
     {$IFDEF USES_NEWFPCX64MM} mormot.core.fpcx64mm, {$ENDIF}
    {$ENDIF}
   {$ENDIF}
 {$ENDIF}
 {$IFDEF WINDOWS}
  {$IFDEF USES_SHAREMEM}
   // sharemem, FastMM5,
   {$IFDEF USES_SYNFPCX64MM} synfpcx64mm, {$ENDIF}
   {$IFDEF USES_NEWFPCX64MM} mormot.core.fpcx64mm, {$ENDIF}
  {$ENDIF}
 {$ENDIF}
 sysutils;

var
 ShareMemStorage : record
  DefMM: TMemoryManager;     // Default  Memory Manager
  CurMM: TMemoryManager;     // Current  Memory Manager defined by user
  ExtMM: TMemoryManager;     // Extended Memory Manager to control memory usage
  ExtMM_Enabled:Boolean;     // ExtMM usage is Enabled
  ExtMM_Allocated: SizeInt;  // ExtMM allocated memory counter
  MainModuleHandle: Pointer; // Uses to get CommonMemoryManager
  GetCommonMemoryManager: procedure(var aMemMgr: TMemoryManager);
 end;

implementation

procedure LockedAdd(var Target:SizeInt; Value:SizeInt); inline;
begin
 {$IFDEF CPU64}
 if (Value<>0) then System.InterlockedExchangeAdd64(Target,Value);
 {$ELSE}
 if (Value<>0) then System.InterlockedExchangeAdd(Target,Value);
 {$ENDIF}
end;

function extMemSize(p:Pointer):SizeInt; inline;
begin
 if Assigned(p)
 then Result:=ShareMemStorage.CurMM.MemSize(p)
 else Result:=0;
end;

function extGetmem(Size:PtrUInt):Pointer;
begin
 Result:=ShareMemStorage.CurMM.Getmem(Size);
 LockedAdd(ShareMemStorage.ExtMM_Allocated,extMemSize(Result));
end;

function extFreemem(p:Pointer):PtrUInt;
begin
 LockedAdd(ShareMemStorage.ExtMM_Allocated,-extMemSize(p));
 Result:=ShareMemStorage.CurMM.Freemem(p);
end;

function extFreememSize(p:Pointer; Size:PtrUInt):PtrUInt;
begin
 LockedAdd(ShareMemStorage.ExtMM_Allocated,-extMemSize(p));
 Result:=ShareMemStorage.CurMM.FreememSize(p,Size);
end;

function extAllocMem(Size:PtrUInt):Pointer;
begin
 Result:=ShareMemStorage.CurMM.AllocMem(Size);
 LockedAdd(ShareMemStorage.ExtMM_Allocated,extMemSize(Result));
end;

function extReAllocMem(var p:pointer;Size:ptruint):Pointer;
var n1,n2:SizeInt;
begin
 n1:=extMemSize(p);
 Result:=ShareMemStorage.CurMM.ReAllocMem(p,Size);
 n2:=extMemSize(p);
 LockedAdd(ShareMemStorage.ExtMM_Allocated,n2-n1);
end;

procedure Init_crw_sharemem;
begin
 with ShareMemStorage do begin
  MainModuleHandle:=nil;
  GetMemoryManager(DefMM);
  GetCommonMemoryManager:=@GetMemoryManager;
  ExtMM_Enabled:={$IFDEF USES_EXTMM} true {$ELSE} false {$ENDIF};
  {$IFDEF USES_SHAREMEM}
   {$IFDEF UNIX}
    {$IFDEF IS_DLL_UNIT}
     MainModuleHandle:=dlopen(nil, RTLD_LAZY);
     GetCommonMemoryManager:=dlsym(MainModuleHandle,'GetMemoryManager');
     GetCommonMemoryManager(CurMM);
     SetMemoryManager(CurMM);
    {$ENDIF}
   {$ELSE}
    {$IFDEF FPC}
    // Uses fpcx64mm
    {$ELSE}
    if IsLibrary // Sharing MemoryManager
    then FastMM_AttemptToUseSharedMemoryManager
    else FastMM_ShareMemoryManager;
    {$ENDIF}
   {$ENDIF}
  {$ENDIF}
  GetMemoryManager(CurMM);
  GetMemoryManager(ExtMM);
  if ExtMM_Enabled then begin
   ExtMM.Getmem:=extGetmem;
   ExtMM.Freemem:=extFreemem;
   ExtMM.FreememSize:=extFreememSize;
   ExtMM.AllocMem:=extAllocMem;
   ExtMM.ReAllocMem:=extReAllocMem;
   SetMemoryManager(ExtMM);
  end;
 end;
end;

procedure Free_crw_sharemem;
begin
 with ShareMemStorage do begin
  if ExtMM_Enabled
  then SetMemoryManager(CurMM);
  {$IFDEF USES_SHAREMEM}
   {$IFDEF UNIX}
    {$IFDEF IS_DLL_UNIT}
     SetMemoryManager(DefMM);
    {$ENDIF}
   {$ENDIF}
  {$ENDIF}
 end;
end;

initialization

 Init_crw_sharemem;

finalization

 Free_crw_sharemem;

end.

//////////////
// END OF FILE
//////////////

